# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

"""Backend of the problem setup step."""

import json
import time

from pathlib import Path
from typing import List, Optional

from ansys.optislang.core import Optislang, logging
from ansys.saf.glow.solution import StepModel, StepSpec, long_running, transaction, FileReference

from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.solution.problem_setup_step import ProblemSetupStep
from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.utilities.common_functions import read_log_file


class MonitoringStep(StepModel):
    """Step model of the monitoring step."""

    # Parameters ------------------------------------------------------------------------------------------------------

    # Frontend persistence
    selected_actor_from_treeview: Optional[str] = None
    selected_command: Optional[str] = None
    selected_actor_from_command: Optional[str] = None
    commands_locked: bool = False
    auto_update_frequency: float = 2000
    auto_update_activated: bool = True
    actor_uid: Optional[str] = None
    project_command_execution_status: dict = {"alert-message": "", "alert-color": "info"}
    actor_command_execution_status: dict = {"alert-message": "", "alert-color": "info"}
    project_btn_group_options: List = [
                                {
                                    "icon": "fas fa-play",
                                    "tooltip": "Restart optiSLang project.",
                                    "value": "restart",
                                    "id": {
                                        "type":"action-button",
                                        "action":"restart"
                                    }
                                },
                                {
                                    "icon":"fa fa-hand-paper",
                                    "tooltip": "Stop optiSLang project gently.",
                                    "value": "stop_gently",
                                    "id": {
                                        "type":"action-button",
                                        "action":"stop_gently"
                                    }
                                },
                                {
                                    "icon": "fas fa-stop",
                                    "tooltip":"Stop optiSLang project.",
                                    "value": "stop",
                                    "id": {
                                        "type": "action-button",
                                        "action":"stop"
                                        }
                                    },
                                {
                                    "icon": "fas fa-fast-backward",
                                    "tooltip": "Reset optiSLang project.",
                                    "value": "reset",
                                    "id": {
                                        "type":"action-button",
                                        "action":"reset"
                                        }
                                    },
                                {
                                    "icon": "fas fa-power-off",
                                    "tooltip": "Shutdown optiSLang project.",
                                    "value": "shutdown",
                                    "id": {
                                        "type":"action-button",
                                        "action":"shutdown"
                                    }
                                },
                            ]
    actor_btn_group_options: List = [
                            {
                                "icon": "fas fa-play",
                                "tooltip": "Restart node.",
                                "value": "restart",
                                "id": {
                                    "type":"action-button",
                                    "action":"restart"
                                }
                            },

                            {
                                "icon":"fa fa-hand-paper",
                                "tooltip": "Stop node gently.",
                                "value": "stop_gently",
                                "id": {
                                    "type":"action-button",
                                    "action":"stop_gently"
                                }
                            },

                            {
                                "icon": "fas fa-stop",
                                "tooltip":"Stop node.",
                                "value": "stop",
                                "id": {
                                    "type": "action-button",
                                    "action":"stop"
                                }
                            },

                            {
                                "icon": "fas fa-fast-backward",
                                "tooltip": "Reset node.",
                                "value": "reset",
                                "id": {
                                    "type":"action-button",
                                    "action":"reset"
                                }
                            },
                        ]

    # Backend data model
    results_files: dict = {}
    project_state: str = "NOT STARTED"
    osl_project_states: List = [
        "IDLE",
        "PROCESSING",
        "PAUSED",
        "PAUSE_REQUESTED",
        "STOPPED",
        "STOP_REQUESTED",
        "GENTLY_STOPPED",
        "GENTLE_STOP_REQUESTED",
        "FINISHED"
    ]
    osl_actor_states: List = [
        "Idle",
        "Succeeded",
        "Failed",
        "Running",
        "Aborted",
        "Predecessor failed",
        "Skipped",
        "Incomplete",
        "Processing done",
        "Stopped",
        "Gently stopped"
    ]
    optislang_logs: List = []
    command_timeout: int = 30

    # File storage ----------------------------------------------------------------------------------------------------

    # Output
    optislang_log_file: FileReference = FileReference("Problem_Setup/pyoptislang.log")
    project_status_info_file: FileReference = FileReference("Problem_Setup/project_status_info.json")
    actors_info_file: FileReference = FileReference("Problem_Setup/actors_info.json")
    actors_status_info_file: FileReference = FileReference("Problem_Setup/actors_status_info.json")

    # Methods ---------------------------------------------------------------------------------------------------------

    @transaction(
        problem_setup_step=StepSpec(
            download=[
                "tcp_server_host",
                "tcp_server_port",
                "project_tree",
                "optislang_log_level",
            ]
        ),
        self=StepSpec(
            upload=[
                "project_state",
                "project_status_info_file",
                "actors_info_file",
                "actors_status_info_file",
                "optislang_logs",
                "optislang_log_file",
                "results_files",
            ],
        )
    )
    @long_running
    def upload_project_data(self, problem_setup_step: ProblemSetupStep) -> None:
        """Monitor the progress of the optiSLang project and continuously upload project data."""
        # Creat monitoring directory
        Path(self.optislang_log_file.path).parent.mkdir(parents=True, exist_ok=True)

        # Connect to optiSLang instance.
        osl = Optislang(
            host=problem_setup_step.tcp_server_host,
            port=problem_setup_step.tcp_server_port,
            loglevel=problem_setup_step.optislang_log_level,
            shutdown_on_finished=True
        )

        # Configure logging.
        osl_logger = logging.OslLogger(
            loglevel=problem_setup_step.optislang_log_level,
            log_to_file=True,
            logfile_name=self.optislang_log_file.path,
            log_to_stdout=True,
        )
        osl.__logger = osl_logger.add_instance_logger(osl.name, osl, problem_setup_step.optislang_log_level)

        # Monitor project state and upload data.
        while True:
            # Get project state
            self.project_state = osl.project.get_status()
            osl.log.info(f"Analysis status: {self.project_state}")
            # Get project status info
            with open(self.project_status_info_file.path, "w") as json_file: json.dump(osl.get_osl_server().get_full_project_status_info(), json_file)
            # Read pyoptislang logs
            self.optislang_logs = read_log_file(self.optislang_log_file.path)
            # Get actor status info
            actors_info, actors_status_info = {}, {}
            for node_info in problem_setup_step.project_tree:
                if node_info["uid"] == osl.project.root_system.uid:
                    node = osl.project.root_system
                else:
                    node = osl.project.root_system.find_node_by_uid(node_info["uid"], search_depth=-1)
                actors_info[node.uid] = osl.get_osl_server().get_actor_info(node.uid)
                if node.get_states_ids():
                    actors_status_info[node.uid] = []
                    for hid in node.get_states_ids():
                        actors_status_info[node.uid].append(
                            osl.get_osl_server().get_actor_status_info(node.uid, hid)
                        )
            with open(self.actors_info_file.path, "w") as json_file: json.dump(actors_info, json_file)
            with open(self.actors_status_info_file.path, "w") as json_file: json.dump(actors_status_info, json_file)
            # Upload fields
            self.transaction.upload(["project_state"])
            self.transaction.upload(["project_status_info_file"])
            self.transaction.upload(["actors_info_file"])
            self.transaction.upload(["actors_status_info_file"])
            self.transaction.upload(["optislang_logs"])
            if self.project_state == "FINISHED":
                break
            time.sleep(3) # Waiting 3 sec before pulling new data. The frequency might be adjusted in the future.

        # Close connection with optiSLang server.
        osl.dispose()

    @transaction(
        problem_setup_step=StepSpec(
            download=[
                "tcp_server_host",
                "tcp_server_port",
                "project_tree",
            ]
        ),
        self=StepSpec(
            download=[
                "command_timeout",
                "selected_actor_from_command",
                "selected_command",
            ],
            upload=["actor_uid"],
        )
    )
    def control_node_state(self, problem_setup_step: ProblemSetupStep) -> None:
        """Update the state of root or actor node based on the selected command in the UI."""
        osl = Optislang(
                host=problem_setup_step.tcp_server_host,
                port=problem_setup_step.tcp_server_port,
                shutdown_on_finished=False
            )
        if not self.selected_actor_from_command == "shutdown":
            if self.selected_actor_from_command == osl.project.root_system.uid:
                node = osl.project.root_system
                self.actor_uid = None
            else:
                node = osl.project.root_system.find_node_by_uid(self.selected_actor_from_command, search_depth=-1)
                self.actor_uid = node
        else:
            node = osl.project.root_system

        status = node.control(self.selected_command, wait_for_completion=True, timeout=self.command_timeout)

        if not status:
            raise Exception(f"{self.selected_command.replace('_', ' ').title()} command against node {node.get_name()} failed.")

        osl.dispose()



# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

"""Backend of the problem setup step."""

import json
import os
import platform
import tempfile

from pathlib import Path
from typing import List, Optional

from ansys.optislang.core import Optislang, utils
from ansys.saf.glow.solution import FileReference, FileGroupReference, StepModel, StepSpec, long_running, transaction
from ansys.solutions.optislang.frontend_components.project_properties import ProjectProperties, write_properties_file, apply_placeholders_to_properties_file

from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.utilities.common_functions import get_treeview_items_from_project_tree


class ProblemSetupStep(StepModel):
    """Step model of the problem setup step."""

    # Parameters ------------------------------------------------------------------------------------------------------

    # Frontend persistence
    ansys_ecosystem_ready: bool = False
    project_initialized: bool = False
    analysis_locked: bool = True
    project_locked: bool = False

    # Backend data model
    tcp_server_host: Optional[str] = None
    tcp_server_port: Optional[int] = None
    ansys_ecosystem: dict = {
        "optislang": {
            "authorized_versions": [],
            "installed_versions": [],
            "compatible_versions": [],
            "selected_version": None,
            "alert_message": "optiSLang install not checked.",
            "alert_color": "warning",
            "alias": "optiSLang",
        }
    }
    placeholders: dict = {}
    registered_files: List = []
    settings: dict = {}
    parameter_manager: dict = {}
    criteria: dict = {}
    ui_placeholders: dict = {}
    app_metadata: dict = {}
    project_tree: list = []
    treeview_items: list = [
        {
            "key": "problem_setup_step",
            "text": "Problem Setup",
            "depth": 0,
            "uid": None
        },
    ]
    optislang_log_level: str = "INFO"

    # File storage ----------------------------------------------------------------------------------------------------

    # Inputs
    project_file: FileReference = FileReference("Problem_Setup/{{ cookiecutter.__optiSLang_application_archive_stem }}.opf")
    properties_file: FileReference = FileReference("Problem_Setup/{{ cookiecutter.__optiSLang_application_archive_stem }}.json")
    metadata_file: FileReference = FileReference("Problem_Setup/metadata.json")
    input_files: FileGroupReference = FileGroupReference("Problem_Setup/Input_Files/*.*")
    # If folder doesn't exist, it will be created later
    upload_directory: str = os.path.join(tempfile.gettempdir(), "GLOW")

    # Outputs
    working_properties_file: FileReference = FileReference("Problem_Setup/working_properties_file.json")

    # Methods ---------------------------------------------------------------------------------------------------------

    @transaction(
        self=StepSpec(
            upload=[
                "project_file",
                "properties_file",
                "metadata_file",
            ]
        )
    )
    @long_running
    def upload_bulk_files_to_project_directory(self) -> None:
        """Upload bulk files to project directory."""
        original_project_file = Path(__file__).parent.absolute().parent / "model" / "assets" / "{{ cookiecutter.__optiSLang_application_archive_stem }}.opf"
        self.project_file.write_bytes(original_project_file.read_bytes())

        original_properties_file = (
            Path(__file__).parent.absolute().parent / "model" / "assets" / "{{ cookiecutter.__optiSLang_application_archive_stem }}.json"
        )
        self.properties_file.write_bytes(original_properties_file.read_bytes())

        original_metadata_file = Path(__file__).parent.absolute().parent / "model" / "assets" / "metadata.json"
        self.metadata_file.write_bytes(original_metadata_file.read_bytes())

    @transaction(
        self=StepSpec(
            download=["properties_file"],
            upload=["placeholders", "registered_files", "settings", "parameter_manager", "criteria"]
        )
    )
    @long_running
    def get_default_placeholder_values(self):
        """Get placeholder values and definitions using the ProjectProperties class."""
        pp = ProjectProperties()
        pp.read_file(self.properties_file.path)
        self.placeholders = pp._placeholders
        self.registered_files = pp._registered_files
        self.settings = pp._settings
        self.parameter_manager = pp._parameter_manager
        self.criteria = pp._criteria

    @transaction(
        self=StepSpec(
            download=["properties_file", "ui_placeholders"],
            upload=[
                "working_properties_file",
                "placeholders",
                "registered_files",
                "settings",
                "parameter_manager",
                "criteria",
            ],
        )
    )
    def write_updated_properties_file(self) -> None:
        """Write updated optiSLang project properties file."""
        properties = apply_placeholders_to_properties_file(self.ui_placeholders, self.properties_file.path)
        self.placeholders = properties["placeholders"]
        self.registered_files = properties["registered_files"]
        self.settings = properties["settings"]
        self.parameter_manager = properties["parameter_manager"]
        self.criteria = properties["criteria"]
        write_properties_file(properties, Path(self.working_properties_file.path))

    @transaction(
        self=StepSpec(
            download=["properties_file", "ui_placeholders"],
            upload=["placeholders", "registered_files", "settings", "parameter_manager", "criteria"],
        )
    )
    def update_osl_placeholders_with_ui_values(self) -> None:
        """Update placeholders with values selected by the user in the UI."""
        properties = apply_placeholders_to_properties_file(self.ui_placeholders, self.properties_file.path)
        self.placeholders = properties["placeholders"]
        self.registered_files = properties["registered_files"]
        self.settings = properties["settings"]
        self.parameter_manager = properties["parameter_manager"]
        self.criteria = properties["criteria"]

    @transaction(
        self=StepSpec(
            upload=["ansys_ecosystem", "ansys_ecosystem_ready"],
        )
    )
    def check_ansys_ecosystem(self) -> None:
        """Check if Ansys Products are installed and if the appropriate versions are available."""
        self.ansys_ecosystem_ready = True

        # Collect optiSLang installations
        self.ansys_ecosystem["optislang"]["installed_versions"] = sorted(list(dict(utils.find_all_osl_exec()).keys()))
        if len(self.ansys_ecosystem["optislang"]["authorized_versions"]) > 0:
            self.ansys_ecosystem["optislang"]["compatible_versions"] = [
                product_version
                for product_version in self.ansys_ecosystem["optislang"]["installed_versions"]
                if product_version in self.ansys_ecosystem["optislang"]["authorized_versions"]
            ]
        else:
            self.ansys_ecosystem["optislang"]["compatible_versions"] = self.ansys_ecosystem["optislang"]["installed_versions"]

        # Check ecosystem
        for product_name in self.ansys_ecosystem.keys():
            if len(self.ansys_ecosystem[product_name]["installed_versions"]) == 0:
                alert_message = f"No installation of {product_name.title()} found in the machine {platform.node()}."
                alert_color = "danger"
                self.ansys_ecosystem_ready = False
            elif len(self.ansys_ecosystem[product_name]["compatible_versions"]) == 0:
                alert_message = (
                    f"None of the authorized versions of {product_name.title()} "
                    f"is installed in the machine {platform.node()}.\n"
                )
                alert_message += "At least one of these versions is required:"
                for authorized_version in self.ansys_ecosystem[product_name]["authorized_versions"]:
                    self.ansys_ecosystem[product_name][
                        "alert_message"
                    ] += f" {authorized_version}"
                alert_message += "."
                alert_color = "danger"
                self.ansys_ecosystem_ready = False
            else:
                self.ansys_ecosystem[product_name]["selected_version"] = self.ansys_ecosystem[product_name][
                    "compatible_versions"
                ][
                    -1
                ]  # Latest
                alert_message = f"{product_name.title()} install detected. Compatible versions are:"
                for compatible_version in self.ansys_ecosystem[product_name]["compatible_versions"]:
                    alert_message += f" {compatible_version}"
                alert_message += ".\n"
                alert_message += "Selected version is %s." % (self.ansys_ecosystem[product_name]["selected_version"])
                alert_color = "success"
            self.ansys_ecosystem[product_name]["alert_message"] = alert_message
            self.ansys_ecosystem[product_name]["alert_color"] = alert_color

    @transaction(
        self=StepSpec(
            download=["metadata_file"],
            upload=["app_metadata"]
        )
    )
    @long_running
    def get_app_metadata(self) -> None:
        """Read OWA metadata file."""
        with open(self.metadata_file.path) as f:
            self.app_metadata = json.load(f)

    @transaction(
        self=StepSpec(
            download=[
                "project_file",
            ],
            upload=[
                "project_tree",
                "treeview_items"
            ],
        )
    )
    @long_running
    def get_project_tree(self) -> None:
        """Get the project tree of the optiSLang project."""
        # Start optiSLang instance.
        osl = Optislang(
            project_path=self.project_file.path,
            reset=True,
            shutdown_on_finished=True,
            ini_timeout=300,
        )

        # Get project tree
        self.project_tree = osl.project._get_project_tree()

        # Update treeview items
        self.treeview_items = get_treeview_items_from_project_tree(self.project_tree)

    @transaction(
        self=StepSpec(
            download=[
                "input_files",
                "project_file",
                "working_properties_file",
                "optislang_log_level"
            ],
            upload=[
                "tcp_server_host",
                "tcp_server_port",
            ],
        )
    )
    def start(self) -> None:
        """Start optiSLang and run the project."""
        # Start optiSLang instance.
        osl = Optislang(
            project_path=self.project_file.path,
            reset=True,
            loglevel=self.optislang_log_level,
            shutdown_on_finished=False,
            import_project_properties_file=self.working_properties_file.path,
            ini_timeout=300,  # might need to be adjusted depending on the hardware
        )

        # Get server host
        self.tcp_server_host = osl.get_osl_server().get_host()

        # Get server port
        server_info = osl.get_osl_server().get_server_info()
        self.tcp_server_port = server_info["server"]["server_port"]

        # Start optiSLang project
        osl.start(wait_for_started=True, wait_for_finished=False)

        # Close connection with optiSLang server
        osl.dispose()



# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

"""Common functions."""

import dash_bootstrap_components as dbc
import re

from dash_extensions.enrich import html
from pathlib import Path
from typing import Any, Iterable, Union


MONITORING_TABS = [
    {
        "label": "Project Summary",
        "tab_id": "project_summary_tab",
        "is_root": True,
        "is_system": False,
        "is_actor": False,
    },
    {"label": "Summary", "tab_id": "summary_tab", "is_root": False, "is_system": True, "is_actor": True},
    {"label": "Result Files", "tab_id": "result_files_tab", "is_root": True, "is_system": False, "is_actor": False},
    {"label": "Scenery", "tab_id": "scenery_tab", "is_root": True, "is_system": False, "is_actor": False},
    {"label": "Design Table", "tab_id": "design_table_tab", "is_root": True, "is_system": True, "is_actor": False},
    {"label": "Visualization", "tab_id": "visualization_tab", "is_root": True, "is_system": True, "is_actor": False},
    {"label": "Status Overview", "tab_id": "status_overview_tab", "is_root": True, "is_system": True, "is_actor": True},
]


PROJECT_STATES = {
    "NOT STARTED": {
        "alert": "optiSLang project not started.",
        "color": "warning"
    },
    "IDLE": {
        "alert": "optiSLang project is pending.",
        "color": "warning"
    },
    "PROCESSING": {
        "alert": "optiSLang project in progress.",
        "color": "primary"
    },
    "PAUSED": {
        "alert": "optiSLang project paused.",
        "color": "warning"
    },
    "PAUSE_REQUESTED": {
        "alert": "optiSLang project requested to pause.",
        "color": "warning"
    },
    "STOPPED": {
        "alert": "optiSLang project stopped.",
        "color": "warning"
    },
    "STOP_REQUESTED": {
        "alert": "optiSLang project requetsed to stop.",
        "color": "warning"
    },
    "GENTLY_STOPPED": {
        "alert": "optiSLang project gently stopped.",
        "color": "warning"
    },
    "GENTLE_STOP_REQUESTED": {
        "alert": "optiSLang project requested to gently stop.",
        "color": "warning"
    },
    "FINISHED": {
        "alert": "optiSLang project completed successfully.",
        "color": "success"
    },
}


def get_treeview_items_from_project_tree(project_tree: list) -> list:

    treeview_items = [
        {
            "key": "problem_setup_step",
            "text": "Problem Setup",
            "depth": 0,
            "uid": None,
        },
    ]

    for node in project_tree:
        if node["kind"] == "system":
            treeview_items.append(
                {
                    "key": f'{node["name"].lower()}_{node["uid"]}_toggle',
                    "text": node["name"],
                    "depth": node["level"],
                    "uid": node["uid"],
                }
            )
        treeview_items.append(
            {
                "key": f'{node["name"].lower()}_{node["uid"]}',
                "text": node["name"],
                "depth": node["level"] + 1 if node["kind"] == "system" else node["level"],
                "uid": node["uid"],
            }
        )

    return treeview_items


def check_empty_strings(lst) -> bool:
    for sublist in lst:
        for item in sublist:
            if not item.strip():  # Using strip() to remove leading/trailing whitespaces
                return False
    return True


def sort_dict_by_ordered_keys(dictionary: dict, list_of_keys: list) -> dict:
    return {key: dictionary[key] for key in list_of_keys if key in dictionary}


def sorted_nicely(l) -> Iterable:
    """Sort the given iterable in the way that humans expect.
    Reference: https://stackoverflow.com/questions/2669059/how-to-sort-alpha-numeric-set-in-python
    """

    convert = lambda text: int(text) if text.isdigit() else text
    alphanum_key = lambda key: [convert(c) for c in re.split("([0-9]+)", key)]
    return sorted(l, key=alphanum_key)


def convert_microseconds(duration: int) -> str:
    """Convert a time duration in microseconds into a string with format: WWh XXmin YYs ZZms."""

    milliseconds = duration // 1000
    microseconds = duration % 1000
    seconds = milliseconds // 1000
    milliseconds = milliseconds % 1000
    minutes = seconds // 60
    seconds = seconds % 60
    hours = minutes // 60
    minutes = minutes % 60

    return f"{int(hours):02}h {int(minutes):02}m {int(seconds):02}s {int(milliseconds):03}ms"


def remove_key_from_dictionaries(dictionaries: list, key: str) -> list:

    for dictionary in dictionaries:
        if key in dictionary:
            del dictionary[key]

    return dictionaries


def update_list_of_tabs(node_info: dict) -> list:

    list_of_tabs = []

    for tab_info in MONITORING_TABS:
        if (
            node_info["is_root"]
            and tab_info["is_root"]
            or not node_info["is_root"]
            and node_info["kind"] == "system"
            and tab_info["is_system"]
            or not node_info["is_root"]
            and node_info["kind"] == "actor"
            and tab_info["is_actor"]
        ):
            list_of_tabs.append(
                dbc.Tab(
                    label=tab_info["label"],
                    tab_id=tab_info["tab_id"],
                    label_style={
                        "color": "#000000",
                        "text-color": "#000000",
                    },
                    active_label_style={
                        "color": "#FFFFFF",
                        "text-color": "#000000",
                        "background-color": "#000000",
                        "border-style": "solid",
                        "border-color": "#000000",
                    },
                )
            )

    return list_of_tabs


def extract_dict_by_key(dictionaries: list, key: str, value: Any, expect_unique: bool = False, return_index: bool = True) -> Union[None, dict]:
    """Given a list of dictionaries, return the ones matching specific key and value."""

    matches, indexes = [], []

    for (index, dictionary) in enumerate(dictionaries):
        if dictionary.get(key) == value:
            matches.append(dictionary)
            indexes.append(index)

    if expect_unique:
        if len(matches) == 0:
            raise Exception(f"No matching dictionaries with key {key} and value {value}.")
        elif len(matches) == 1:
            result_m = matches[0]
            result_i = indexes[0]
        else:
            raise Exception(f"Multiple matches found with key {key} and value {value}.")
    else:
        result_m = matches
        result_i = indexes

    return (result_m, result_i) if return_index else result_m


def read_log_file(log_file: Union[Path, str]) -> list:

    logs = []

    log_file = Path(log_file)

    if log_file.exists():
        with open(log_file, "r") as file:
            for line in file:
                if len(line.strip().rstrip("\n").replace(" ", "")) > 0:
                    logs.append(line.strip().rstrip("\n"))
    else:
        raise FileNotFoundError(f"Unable to find log file {log_file}.")

    return logs


def update_placeholders(ui_values: list, placeholders: dict) -> dict:
    """Information needed."""

    updated_dict = {}
    for row in ui_values[1:]:
        parameter_name = row["props"]["children"][0]["props"]["children"]["props"]["children"]
        input_value = row["props"]["children"][1]["props"]["children"]["props"]["value"]

        placeholder_values = placeholders.get("placeholder_values")
        if parameter_name in placeholder_values:
            updated_dict[parameter_name] = input_value
    return updated_dict


def update_alerts(problem_setup_step, monitoring_step) -> list:
    """Update all Alerts."""

    alerts = []

    # Product version alerts
    for product_name, product_data in problem_setup_step.ansys_ecosystem.items():
        alerts.append(
            html.Div(
                [
                    dbc.Button(
                        f"{product_data['alias']} Version",
                        id=f"popover_{product_name}_version_target",
                        disabled=False,
                        color=product_data["alert_color"],
                        n_clicks=0,
                    ),
                    dbc.Popover(
                        [
                            dbc.PopoverBody(product_data["alert_message"]),
                        ],
                        id=f"popover_{product_name}_version",
                        target=f"popover_{product_name}_version_target",
                        placement="top",
                        is_open=False,
                    ),
                ]
            ),
        )

    # optiSLang solve alert
    if monitoring_step.project_state in PROJECT_STATES.keys():
        solve_message, solve_color = PROJECT_STATES[monitoring_step.project_state]["alert"], PROJECT_STATES[monitoring_step.project_state]["color"]
    else:
        raise ValueError(f"Unknown optiSLang state: {monitoring_step.project_state}.")

    alerts.append(
        html.Div(
            [
                dbc.Button(
                    "optiSLang Solve",
                    id="popover_optislang_solve_target",
                    disabled=False,
                    color=solve_color,
                    n_clicks=0,
                ),
                dbc.Popover(
                    [
                        dbc.PopoverBody(solve_message),
                    ],
                    id="popover_optislang_solve",
                    target="popover_optislang_solve_target",
                    placement="top",
                    is_open=False,
                ),
            ]
        ),
    )

    return alerts



# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

"""Frontend of the problem setup step."""

import dash_bootstrap_components as dbc
import os
import time

from dash.exceptions import PreventUpdate
from dash_extensions.enrich import Input, Output, State, dcc, html, ALL, MATCH, no_update, callback_context

from ansys.saf.glow.client.dashclient import DashClient, callback
from ansys.saf.glow.core.method_status import MethodStatus
from ansys.solutions.optislang.frontend_components.load_sections import to_dash_sections, update_designs_to_dash_section

from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.solution.definition import {{ cookiecutter.__solution_definition_name }}
from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.solution.problem_setup_step import ProblemSetupStep
from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.solution.monitoring_step import MonitoringStep
from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.utilities.common_functions import check_empty_strings, update_alerts


def layout(problem_setup_step: ProblemSetupStep, monitoring_step: MonitoringStep) -> html.Div:
    """Layout of the problem setup step."""

    project_properties_sections = to_dash_sections(
            problem_setup_step.placeholders, problem_setup_step.registered_files, problem_setup_step.project_locked
        )

    return html.Div(
        [
            # Header
            html.H1(
                problem_setup_step.app_metadata["title"].strip().title(),
                className="display-3",
                style={"font-size": "45px"},
            ),
            html.P(
                problem_setup_step.app_metadata["description"].strip(),
                className="lead",
                style={"font-size": "25px"},
            ),
            html.Hr(className="my-2"),
            html.Br(),
            # Alerts
            dbc.Row(
                id="project-locked-alert",
                children=[
                    dbc.Alert(
                        "The selected inputs are locked. To set new inputs, create a new project.",
                        color="info",
                    )
                ],
                style={"display": "block"} if problem_setup_step.project_locked else {"display": "none"},
            ),
            dbc.Row(
                [
                    dbc.Col(
                        [
                            dbc.Stack(
                                update_alerts(problem_setup_step, monitoring_step),
                                id="alert_messages",
                                direction="horizontal",
                                gap=3,
                            ),
                        ],
                        width=12,
                    )
                ]
            ),
            html.Br(),
            # Input form
            html.H1("Input Form", style={"font-size": "20px"}),
            html.Hr(className="my-2"),
            dbc.Row(id="osl-dash-sections",children=project_properties_sections),
            html.Br(),
            # Start analysis
            html.H1("Start Analysis", style={"font-size": "20px"}),
            html.Hr(className="my-2"),
            dbc.Row(
                [
                    html.Div(
                        [
                            dbc.Button(
                                html.I(className="fas fa-play", style={"display": "inline-block"}),
                                id="start_analysis",
                                disabled=problem_setup_step.analysis_locked,
                                style = {
                                    "display": "flex",
                                    "justify-content": "center",
                                    "align-items": "center",
                                    "fontSize": "150%",
                                    "background-color": "rgba(0, 0, 0, 1)",
                                    "border-color": "rgba(0, 0, 0, 1)",
                                    "height": "30px",
                                }
                            ),
                            dbc.Tooltip(
                                "Start optiSLang project execution.",
                                target="start_analysis",
                            ),
                        ],
                        className="d-grid gap-2 col-5 mx-auto",
                    ),
                    dcc.Loading(
                        type="circle",
                        fullscreen=True,
                        color="#ffb71b",
                        style={
                            "background-color": "rgba(55, 58, 54, 0.1)",
                        },
                        children=html.Div(id="wait_start_analysis"),
                    ),
                    dcc.Interval(
                        id="solve_interval_component",
                        interval=1 * 3000,  # in milliseconds
                        n_intervals=0,
                    ),
                    dbc.Button(
                        id="dummy_button_2",
                        style={"display": "none"},
                        n_clicks=0,
                        color="dark"
                    ),
                ]
            ),
        ]
    )


for alert in ["optislang_version", "optislang_solve"]:

    @callback(
        Output(f"popover_{alert}", "is_open"),
        [Input(f"popover_{alert}_target", "n_clicks")],
        [State(f"popover_{alert}", "is_open")],
    )
    def toggle_popover(n_clicks, is_open):
        if n_clicks:
            return not is_open
        return is_open


@callback(
    Output("trigger_treeview_display", "data"),
    Output("navigation_tree", "items"),
    Output("alert_messages", "children"),
    Output("wait_start_analysis", "children"),
    Output("start_analysis", "disabled"),
    Output("osl-dash-sections", "children"),
    Output("project-locked-alert", "style"),
    Input("start_analysis", "n_clicks"),
    State("url", "pathname"),
    prevent_initial_call=True,
)
def start_analysis(n_clicks, pathname):
    """Start optiSLang and run the project."""

    project = DashClient[{{ cookiecutter.__solution_definition_name }}].get_project(pathname)
    problem_setup_step = project.steps.problem_setup_step
    monitoring_step = project.steps.monitoring_step

    if n_clicks:
        ui_data = problem_setup_step.ui_placeholders
        ui_data.update({"start_analysis_requested": True})
        problem_setup_step.ui_placeholders = ui_data
        # Check if Ansys products are available
        project.steps.problem_setup_step.check_ansys_ecosystem()
        if problem_setup_step.ansys_ecosystem_ready:
            # Read project tree and update treeview
            problem_setup_step.get_project_tree()
            # Update project properties file prior to the solve
            problem_setup_step.write_updated_properties_file()
            # Start analysis
            problem_setup_step.start()
            # Lock start analysis and ui data
            problem_setup_step.analysis_locked = True
            problem_setup_step.project_locked = True
            # Start monitoring
            monitoring_step.upload_project_data()
            # Wait until the analysis effectively starts
            while monitoring_step.project_state == "NOT STARTED":
                time.sleep(1)
            return (
                True,
                problem_setup_step.treeview_items,
                update_alerts(problem_setup_step, monitoring_step),
                True,
                True,
                to_dash_sections(
                    problem_setup_step.placeholders, problem_setup_step.registered_files, problem_setup_step.project_locked
                ),
                {"display": "block"},
            )
        else:
            return (
                True,
                problem_setup_step.treeview_items,
                update_alerts(problem_setup_step, monitoring_step),
                True,
                False,
                to_dash_sections(
                    problem_setup_step.placeholders, problem_setup_step.registered_files, problem_setup_step.project_locked
                ),
                {"display": "block"},
            )
    else:
        raise PreventUpdate


@callback(
    Output("alert_messages", "children"),
    Input("solve_interval_component", "n_intervals"),
    Input("start_analysis", "n_clicks"),
    State("url", "pathname"),
)
def update_alert_messages(n_intervals, n_clicks, pathname):
    """Display status badges."""

    project = DashClient[{{ cookiecutter.__solution_definition_name }}].get_project(pathname)
    problem_setup_step = project.steps.problem_setup_step
    monitoring_step = project.steps.monitoring_step

    if problem_setup_step.ansys_ecosystem_ready:
        return update_alerts(problem_setup_step, monitoring_step)
    else:
        raise PreventUpdate


@callback(
    Output("start_analysis", "disabled"),
    Input("dummy_button_2", "n_clicks"),
    State({"container": "placeholders", "placeholder": ALL}, "value"),
    State({"container": "placeholders", "placeholder": ALL}, "id"),
    State({"type": "upload", "placeholder": ALL}, "id"),
    State("dummy_button_2", "n_clicks"),
    State("url", "pathname"),
    prevent_initial_call=True,
)
def initialize_dictionary_of_ui_placeholders(n_clicks, data, ids, input_file_ids, dummy_clicks, pathname):
    """This initializes the dictionary of ui placeholders using only the values in the Placeholders section."""
    project = DashClient[{{ cookiecutter.__solution_definition_name }}].get_project(pathname)
    problem_setup_step = project.steps.problem_setup_step
    designs = {}
    parameters = {}
    if problem_setup_step.ui_placeholders == {}:
        ui_data = problem_setup_step.ui_placeholders
        for index in range(len(data)):
            if ids[index]["placeholder"].startswith("StartDesigns"):
                split_values = ids[index]["placeholder"].split("##")
                values = split_values[1:]
                rowId = values[0].split("#")[1]
                key = values[1].split("#")[1]
                value = values[2].split("#")[1]
                if rowId in designs:
                    designs[rowId].update({key: value})
                else:
                    designs[rowId] = {key: value}
            elif "ParameterManager" in ids[index]["placeholder"]:
                split_values = ids[index]["placeholder"].split("#")
                pm_name = split_values[0]
                key = split_values[1]
                if isinstance(data[index], list) and True in data[index]:
                    parameters[key] = True
                elif isinstance(data[index], list) and False in data[index]:
                    parameters[key] = False
                else:
                    parameters[key] = data[index]
                ui_data.update({pm_name: parameters})
            elif "Bool" in ids[index]["placeholder"] and type(data[index]) == list:
                value = data[index]
                if True in value:
                    new_value = True
                else:
                    new_value = False
                ui_data.update({ids[index]["placeholder"]: new_value})
            else:
                ui_data.update({ids[index]["placeholder"]: data[index]})
            ui_data.update({"StartDesigns": designs})
            ui_data["start_analysis_requested"] = False
        if input_file_ids:
            problem_setup_step.analysis_locked = True
        else:
            problem_setup_step.analysis_locked = False
        for index in range(len(input_file_ids)):
            ui_data.update({input_file_ids[index]["placeholder"]: ""})

        problem_setup_step.ui_placeholders = ui_data
        return problem_setup_step.analysis_locked
    else:
        return no_update


@callback(
    Output({"container": "placeholders", "placeholder": MATCH}, "id"),
    Input({"container": "placeholders", "placeholder": MATCH}, "value"),
    State({"container": "placeholders", "placeholder": MATCH}, "id"),
    State("url", "pathname"),
    prevent_initial_call=True,
)
def update_ui_placeholders(value, id, pathname):
    """This updates the dictionary of ui placeholders each time the ui data changes in the Placeholders section"""
    project = DashClient[{{ cookiecutter.__solution_definition_name }}].get_project(pathname)
    problem_setup_step = project.steps.problem_setup_step
    name = id["placeholder"]
    ui_data = problem_setup_step.ui_placeholders
    if name.startswith("StartDesigns"):
        split_values = id["placeholder"].split("##")
        values = split_values[1:]
        rowId = values[0].split("#")[1]
        key = values[1].split("#")[1]
        ui_data["StartDesigns"][rowId][key] = value
    elif "ParameterManager" in name:
        split_values = id["placeholder"].split("#")
        pm_name = split_values[0]
        key = split_values[1]
        if isinstance(value, list) and True in value:
            value = True
        elif isinstance(value, list) and False in value:
            value = False
        ui_data[pm_name][key] = value
    elif "Bool" in name and type(value) == list:
        if True in value:
            value = True
        else:
            value = False
        ui_data.update({name: value})
    else:
        ui_data.update({name: value})
    problem_setup_step.ui_placeholders = ui_data
    update_state = problem_setup_step.get_method_state("update_osl_placeholders_with_ui_values").status
    while update_state == MethodStatus.Running: # current workaround to avoid raising ConflictError: {"detail":"update_osl_placeholders_with_ui_values is already running"}
        time.sleep(0.1)
    problem_setup_step.update_osl_placeholders_with_ui_values()

    return no_update


@callback(
    Output({"type": "upload", "placeholder": MATCH}, "children"),
    Input({"type": "upload", "placeholder": MATCH}, "isCompleted"),
    State({"type": "upload", "placeholder": MATCH}, "fileNames"),
    State({"type": "upload", "placeholder": MATCH}, "upload_id"),
    State({"type": "upload", "placeholder": MATCH}, "id"),
    State("url", "pathname"),
    prevent_initial_call=True,
)
def upload_input_files_to_project_and_update_ui_placeholders(is_completed, filenames, upload_id, component_id, pathname):
    """This uploads an Input file to the project directory and updates the dictionary of ui placeholders every time
    the ui data changes in the Input files section."""
    project = DashClient[{{ cookiecutter.__solution_definition_name }}].get_project(pathname)
    problem_setup_step = project.steps.problem_setup_step

    name = component_id["placeholder"]
    if is_completed and filenames and len(filenames):
        filename = filenames[0]
        filepath = os.path.join(problem_setup_step.upload_directory, upload_id, filename)
        file_data = open(filepath, mode="rb")

        project.upload_file(f"Problem_Setup/Input_Files/{filename}", file_data)
        ui_data = problem_setup_step.ui_placeholders
        ui_data.update({name: filename})
        problem_setup_step.ui_placeholders = ui_data

        return no_update
    else:
        raise PreventUpdate


@callback(
    Output("start_analysis", "disabled"),
    Input({"type": "upload", "placeholder": ALL}, "isCompleted"),
    State({"type": "upload", "placeholder": ALL}, "fileNames"),
    State("url", "pathname"),
    prevent_initial_call=True,
)
def update_status_of_start_analysis_button(is_completed, filenames, pathname):
    """This enables the start analysis button if a file has been uploaded for all input files."""
    project = DashClient[{{ cookiecutter.__solution_definition_name }}].get_project(pathname)
    problem_setup_step = project.steps.problem_setup_step
    if all(filenames):
        problem_setup_step.analysis_locked = False
    else:
        problem_setup_step.analysis_locked = True
    if filenames == True:
        result = check_empty_strings(filenames)

    return problem_setup_step.analysis_locked


@callback(
    Output("dummy_button_2", "n_clicks"),
    Input("url", "pathname"),
    State("dummy_button_2", "n_clicks"),
)
def on_page_load_initialize_dictionary_of_ui_placeholder_values(pathname, n_clicks):
    return n_clicks + 1


@callback(
    Output("start-designs-view", "children"),
    Input({"type": "button-add", "index": ALL}, "n_clicks"),
    Input({"type": "button-del", "index": ALL}, "n_clicks"),
    State({"type": "button-del", "index": ALL}, "disabled"),
    State({"type": "row", "index": ALL}, "id"),
    State("url", "pathname"),
    prevent_initial_call=True,
)
def update_start_designs_table(n_clicks_add, n_clicks_del, disabled_states, row_id, pathname):
    """This updates the StartDesigns table every time the '+' or '-' button is clicked."""
    project = DashClient[{{ cookiecutter.__solution_definition_name }}].get_project(pathname)
    problem_setup_step = project.steps.problem_setup_step
    ctx = callback_context
    triggered_button = ctx.triggered[0]["prop_id"].split(".")[0]
    if (
        ctx.triggered
        and not all(n_click is None for n_click in n_clicks_del)
        or not all(n_click is None for n_click in n_clicks_add)
    ):
        ui_data = problem_setup_step.ui_placeholders
        designs = ui_data["StartDesigns"]
        button_type = triggered_button.split('"type":"')[1].split('"}')[0]
        row_id = triggered_button.split('"index":')[1].split(",")[0]
        if button_type == "button-add":
            designs = add_row(designs, row_id)
        elif button_type == "button-del":
            designs = delete_row(designs, row_id)

        # Rebuild the dictionary with updated keys
        new_designs = {str(index): value for index, (_, value) in enumerate(designs.items(), start=1)}
        ui_data.update({"StartDesigns": new_designs})

        problem_setup_step.ui_placeholders = ui_data
        update_state = problem_setup_step.get_method_state("update_osl_placeholders_with_ui_values").status
        while update_state == MethodStatus.Running: # current workaround to avoid raising ConflictError: {"detail":"update_osl_placeholders_with_ui_values is already running"}
            time.sleep(0.1)
        problem_setup_step.update_osl_placeholders_with_ui_values()

        return update_designs_to_dash_section(problem_setup_step.placeholders, problem_setup_step.project_locked)
    else:
        raise PreventUpdate


def add_row(designs, id_):
    new_value = designs[id_]  # Value from the target key
    new_designs = {}
    for key, value in designs.items():
        new_designs[key] = value
        if key == id_:
            new_designs["temp"] = new_value
    return new_designs


def delete_row(designs, id_):
    del designs[id_]
    return designs
